@@ -17,15 +17,17 @@
 #include <dirent.h>
 #include <sys/wait.h>
 #include <errno.h>
+#include <libgen.h>
 
 #include "ind.h"
 #include "arg.h"
 
 void
 handledir(int sock, char *path, char *port, char *base, char *args,
-		char *sear, char *ohost, char *chost, char *bhost, int istls)
+		char *sear, char *ohost, char *chost, char *bhost, int istls,
+		char *sel, char *traverse)
 {
-	char *pa, *file, *e, *par, *b;
+	char *pa, *file, *e, *par;
 	struct dirent **dirent;
 	int ndir, i, ret = 0;
 	struct stat st;
@@ -34,49 +36,50 @@ handledir(int sock, char *path, char *port, char *base, char *args,
 	USED(args);
 	USED(sear);
 	USED(bhost);
+	USED(sel);
+	USED(traverse);
 
 	pa = xstrdup(path);
-	e = pa + strlen(pa) - 1;
-	if (e[0] == '/')
+
+	/* Is there any directory below the request? */
+	if (strlen(pa+strlen(base)) > 1) {
+		par = xstrdup(pa+strlen(base));
+		e = strrchr(par, '/');
 		*e = '\0';
-
-	par = xstrdup(pa);
-	b = strrchr(par + strlen(base), '/');
-	if (b != NULL) {
-		*b = '\0';
 		dprintf(sock, "1..\t%s\t%s\t%s\r\n",
-			par + strlen(base), ohost, port);
+			par, ohost, port);
+		free(par);
 	}
-	free(par);
 
-	ndir = scandir(pa[0] ? pa : ".", &dirent, 0, alphasort);
+	ndir = scandir(pa, &dirent, 0, alphasort);
 	if (ndir < 0) {
 		perror("scandir");
 		free(pa);
 		return;
 	} else {
 		for (i = 0; i < ndir && ret >= 0; i++) {
-			if (dirent[i]->d_name[0] == '.') {
-				free(dirent[i]);
+			if (dirent[i]->d_name[0] == '.')
 				continue;
-			}
 
 			type = gettype(dirent[i]->d_name);
-			file = smprintf("%s/%s", pa,
+
+			file = smprintf("%s%s%s",
+					pa,
+					pa[strlen(pa)-1] == '/'? "" : "/",
 					dirent[i]->d_name);
 			if (stat(file, &st) >= 0 && S_ISDIR(st.st_mode))
 				type = gettype("index.gph");
-			e = file + strlen(base);
 			ret = dprintf(sock,
 					"%c%-50.50s %10s %16s\t%s\t%s\t%s\r\n",
 					*type->type,
 					dirent[i]->d_name,
 					humansize(st.st_size),
 					humantime(&(st.st_mtime)),
-					e, ohost, port);
+					file + strlen(base), ohost, port);
 			free(file);
-			free(dirent[i]);
 		}
+		for (i = 0; i < ndir; i++)
+			free(dirent[i]);
 		free(dirent);
 	}
 	dprintf(sock, ".\r\n");
@@ -86,31 +89,36 @@ handledir(int sock, char *path, char *port, char *base, char *args,
 
 void
 handlegph(int sock, char *file, char *port, char *base, char *args,
-		char *sear, char *ohost, char *chost, char *bhost, int istls)
+		char *sear, char *ohost, char *chost, char *bhost, int istls,
+		char *sel, char *traverse)
 {
-	Indexs *act;
+	gphindex *act;
 	int i, ret = 0;
 
 	USED(args);
 	USED(sear);
 	USED(bhost);
+	USED(sel);
+	USED(traverse);
 
-	act = scanfile(file);
+	act = gph_scanfile(file);
 	if (act != NULL) {
-		for (i = 0; i < act->num && ret >= 0; i++) {
-			ret = printelem(sock, act->n[i], file, base, ohost, port);
-			freeelem(act->n[i]);
-			act->n[i] = NULL;
-		}
+		for (i = 0; i < act->num && ret >= 0; i++)
+			ret = gph_printelem(sock, act->n[i], file, base, ohost, port);
 		dprintf(sock, ".\r\n");
 
-		freeindex(act);
+		for (i = 0; i < act->num; i++) {
+			gph_freeelem(act->n[i]);
+			act->n[i] = NULL;
+		}
+		gph_freeindex(act);
 	}
 }
 
 void
 handlebin(int sock, char *file, char *port, char *base, char *args,
-		char *sear, char *ohost, char *chost, char *bhost, int istls)
+		char *sear, char *ohost, char *chost, char *bhost, int istls,
+		char *sel, char *traverse)
 {
 	int fd;
 
@@ -120,6 +128,8 @@ handlebin(int sock, char *file, char *port, char *base, char *args,
 	USED(sear);
 	USED(ohost);
 	USED(bhost);
+	USED(sel);
+	USED(traverse);
 
 	fd = open(file, O_RDONLY);
 	if (fd >= 0) {
@@ -131,34 +141,27 @@ handlebin(int sock, char *file, char *port, char *base, char *args,
 
 void
 handlecgi(int sock, char *file, char *port, char *base, char *args,
-		char *sear, char *ohost, char *chost, char *bhost, int istls)
+		char *sear, char *ohost, char *chost, char *bhost, int istls,
+		char *sel, char *traverse)
 {
-	char *p, *path;
+	char *script, *path, *filec, *scriptc;
 
 	USED(base);
 	USED(port);
 
-	path = xstrdup(file);
-	p = strrchr(path, '/');
-	if (p != NULL)
-		p[1] = '\0';
-	else {
-		free(path);
-		path = NULL;
-	}
-
-	p = strrchr(file, '/');
-	if (p == NULL)
-		p = file;
+	filec = xstrdup(file);
+	scriptc = xstrdup(file);
+	path = dirname(filec);
+	script = basename(scriptc);
 
 	if (sear == NULL)
 		sear = "";
 	if (args == NULL)
 		args = "";
 
-	dup2(sock, 0);
-	dup2(sock, 1);
-	dup2(sock, 2);
+	while (dup2(sock, 0) < 0 && errno == EINTR);
+	while (dup2(sock, 1) < 0 && errno == EINTR);
+	while (dup2(sock, 2) < 0 && errno == EINTR);
 	switch (fork()) {
 	case 0:
 		if (path != NULL) {
@@ -166,10 +169,10 @@ handlecgi(int sock, char *file, char *port, char *base, char *args,
 				break;
 		}
 
-		setcgienviron(p, file, port, base, args, sear, ohost, chost,
-				bhost, istls);
+		setcgienviron(script, file, port, base, args, sear, ohost, chost,
+				bhost, istls, sel, traverse);
 
-		if (execl(file, p, sear, args, ohost, port,
+		if (execl(file, script, sear, args, ohost, port, traverse, sel,
 				(char *)NULL) == -1) {
 			perror("execl");
 			_exit(1);
@@ -179,37 +182,31 @@ handlecgi(int sock, char *file, char *port, char *base, char *args,
 		break;
 	default:
 		wait(NULL);
-		free(path);
+		free(filec);
+		free(scriptc);
 		break;
 	}
 }
 
 void
 handledcgi(int sock, char *file, char *port, char *base, char *args,
-		char *sear, char *ohost, char *chost, char *bhost, int istls)
+		char *sear, char *ohost, char *chost, char *bhost, int istls,
+		char *sel, char *traverse)
 {
 	FILE *fp;
-	char *p, *path, *ln = NULL;
+	char *script, *path, *filec, *scriptc, *ln = NULL;
 	size_t linesiz = 0;
 	ssize_t n;
-	int outpipe[2], ret = 0;
-	Elems *el;
+	int outsocks[2], ret = 0;
+	gphelem *el;
 
-	if (pipe(outpipe) < 0)
+	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, outsocks) < 0)
 		return;
 
-	path = xstrdup(file);
-	p = strrchr(path, '/');
-	if (p != NULL)
-		p[1] = '\0';
-	else {
-		free(path);
-		path = NULL;
-	}
-
-	p = strrchr(file, '/');
-	if (p == NULL)
-		p = file;
+	filec = xstrdup(file);
+	scriptc = xstrdup(file);
+	path = dirname(filec);
+	script = basename(scriptc);
 
 	if (sear == NULL)
 		sear = "";
@@ -220,17 +217,17 @@ handledcgi(int sock, char *file, char *port, char *base, char *args,
 	while (dup2(sock, 2) < 0 && errno == EINTR);
 	switch (fork()) {
 	case 0:
-		while(dup2(outpipe[1], 1) < 0 && errno == EINTR);
-		close(outpipe[0]);
+		while (dup2(outsocks[1], 1) < 0 && errno == EINTR);
+		close(outsocks[0]);
 		if (path != NULL) {
 			if (chdir(path) < 0)
 				break;
 		}
 
-		setcgienviron(p, file, port, base, args, sear, ohost, chost,
-				bhost, istls);
+		setcgienviron(script, file, port, base, args, sear, ohost, chost,
+				bhost, istls, sel, traverse);
 
-		if (execl(file, p, sear, args, ohost, port,
+		if (execl(file, script, sear, args, ohost, port, traverse, sel,
 				(char *)NULL) == -1) {
 			perror("execl");
 			_exit(1);
@@ -240,11 +237,12 @@ handledcgi(int sock, char *file, char *port, char *base, char *args,
 		perror("fork");
 		break;
 	default:
-		while(dup2(sock, 1) < 0 && errno == EINTR);
-		close(outpipe[1]);
+		while (dup2(sock, 1) < 0 && errno == EINTR);
+		close(outsocks[1]);
 
-		if (!(fp = fdopen(outpipe[0], "r"))) {
+		if (!(fp = fdopen(outsocks[0], "r"))) {
 			perror("fdopen");
+			close(outsocks[0]);
 			break;
 		}
 
@@ -252,21 +250,22 @@ handledcgi(int sock, char *file, char *port, char *base, char *args,
 			if (ln[n - 1] == '\n')
 				ln[--n] = '\0';
 
-			el = getadv(ln);
+			el = gph_getadv(ln);
 			if (el == NULL)
 				continue;
 
-			ret = printelem(sock, el, file, base, ohost, port);
-			freeelem(el);
+			ret = gph_printelem(sock, el, file, base, ohost, port);
+			gph_freeelem(el);
 		}
 		if (ferror(fp))
 			perror("getline");
 		dprintf(sock, ".\r\n");
 
 		free(ln);
-		free(path);
 		fclose(fp);
 		wait(NULL);
+		free(filec);
+		free(scriptc);
 		break;
 	}
 }
