@@ -3,6 +3,7 @@
  * by 20h
  */
 
+#include <limits.h>
 #include <unistd.h>
 #include <dirent.h>
 #include <memory.h>
@@ -50,7 +51,7 @@ int glfd = -1;
 int dosyslog = 0;
 int logpriority = LOG_INFO|LOG_DAEMON;
 int loglvl = 47;
-int revlookup = 1;
+int revlookup = 0;
 char *logfile = NULL;
 
 int *listfds = NULL;
@@ -59,7 +60,7 @@ int nlistfds = 0;
 char *argv0;
 char stdbase[] = "/var/gopher";
 char *stdport = "70";
-char *indexf[] = {"/index.gph", "/index.cgi", "/index.dcgi", "/index.bin"};
+char *indexf[] = {"index.gph", "index.cgi", "index.dcgi", "index.bob", "index.bin"};
 char *nocgierr = "3Sorry, execution of the token '%s' was requested, but this "
 	    "is disabled in the server configuration.\tErr"
 	    "\tlocalhost\t70\r\n";
@@ -83,7 +84,7 @@ char *htredir = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
 		"  </body>\n"
 		"</html>\n";
 char *selinval ="3Happy helping ☃ here: "
-		"Sorry, your selector does not start with / or contains '..'. "
+		"Sorry, your selector does contains '..'. "
 		"That's illegal here.\tErr\tlocalhost\t70\r\n.\r\n\r\n";
 
 int
@@ -137,8 +138,9 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 	      char *serverp, int nocgi, int istls)
 {
 	struct stat dir;
-	char recvc[1025], recvb[1025], path[1025], args[1025], argsc[1025],
-		*sear, *c, *sep, *recvbp;
+	char recvc[1025], recvb[1025], path[PATH_MAX+1], args[1025],
+		argsc[1025], traverse[1025], traversec[1025],
+		*sear, *sep, *recvbp, *c;
 	int len = 0, fd, i, maxrecv, pathfallthrough = 0;
 	filetype *type;
 
@@ -147,7 +149,7 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		 * If sticky bit is set on base dir and encryption is not
 		 * used, do not serve.
 		 */
-		if (stat(base, &dir) == -1)
+		if (stat(*base? base : "/", &dir) == -1)
 			return;
 		if (dir.st_mode & S_ISVTX) {
 			dprintf(sock, tlserr, recvc);
@@ -164,6 +166,7 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 	memset(recvc, 0, sizeof(recvc));
 	memset(args, 0, sizeof(args));
 	memset(argsc, 0, sizeof(argsc));
+	memset(traverse, 0, sizeof(argsc));
 
 	maxrecv = sizeof(recvb) - 1;
 	if (rlen > maxrecv || rlen < 0)
@@ -176,6 +179,35 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 	c = strchr(recvb, '\n');
 	if (c)
 		c[0] = '\0';
+
+	memmove(recvc, recvb, rlen+1);
+	/*
+	 * Try to guess if we have some HTTP-like protocol compatibility
+	 * mode.
+	 */
+	if (!nocgi && recvb[0] != '/' && (c = strchr(recvb, ' '))) {
+		*c++ = '\0';
+		if (strchr(recvb, '/'))
+			goto dothegopher;
+		if (snprintf(path, sizeof(path), "%s/%s", base, recvb) <= sizeof(path)) {
+			if (stat(path, &dir) == 0) {
+				if (loglvl & FILES)
+					logentry(clienth, clientp, recvc, "compatibility serving");
+
+				handlecgi(sock, path, port, base, "", "", ohost,
+					clienth, serverh, istls, req, "");
+				return;
+			}
+		}
+	}
+dothegopher:
+
+	/* Do not allow requests including "..". */
+	if (strstr(recvb, "..")) {
+		dprintf(sock, "%s", selinval);
+		return;
+	}
+
 	sear = strchr(recvb, '\t');
 	if (sear != NULL) {
 		*sear++ = '\0';
@@ -188,7 +220,10 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		 * mode. DO NOT ADD ANY OTHER GOPHER+ SUPPORT. GOPHER+ IS
 		 * CRAP.
 		 */
-		if (*sear == '+' || *sear == '$' || *sear == '!' || *sear == '\0') {
+		if ((sear[0] == '+' && sear[1] == '\0')
+				|| (sear[0] == '$' && sear[1] == '\0')
+				|| (sear[0] == '!' && sear[1] == '\0')
+				|| sear[0] == '\0') {
 			if (loglvl & GPLUS)
 				logentry(clienth, clientp, recvb, "gopher+ redirect");
 			dprintf(sock, "+-2\r\n");
@@ -213,42 +248,25 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		return;
 	}
 
-	/*
-	 * FUTURE: Valid cases in gopher we overwrite here, but could be used
-	 * for other geomyidae features:
-	 *
-	 *	request string = "?..." -> "/?..."
-	 *	request string = "" -> "/"
-	 *	request string = "somestring" -> "/somestring"
-	 *
-	 * Be careful, when you consider those special cases to be used
-	 * for some feature. You can easily do good and bad.
-	 *
-	 * Look at printelem() in ind.c for the counterpart of producing
-	 * selectors.
-	 */
-
+	/* Strip off the arguments of req?args style. */
 	c = strchr(recvb, '?');
 	if (c != NULL) {
 		*c++ = '\0';
 		snprintf(args, sizeof(args), "%s", c);
 	}
 
-	if (recvb[0] == '\0') {
-		recvb[0] = '/';
-		recvb[1] = '\0';
+	/* Strip '/' at the end of the request. */
+	for (c = recvb + strlen(recvb) - 1; c >= recvb && c[0] == '/'; c--) {
+		memmove(traversec, traverse, strlen(traverse));
+		/* Prepend to traverse. */
+		snprintf(traverse, sizeof(traverse), "/%s", traversec);
+		c[0] = '\0';
 	}
 
-	/*
-	 * Do not allow requests not beginning with '/' or which contain
-	 * "..".
-	 */
-	if (recvb[0] != '/' || strstr(recvb, "..")){
-		dprintf(sock, "%s", selinval);
-		return;
-	}
-
-	if (snprintf(path, sizeof(path), "%s%s", base, recvb) > sizeof(path)) {
+	/* path is now always at least '/' */
+	if (snprintf(path, sizeof(path), "%s%s%s", base,
+	    (*recvb != '/')? "/" : "",
+	    recvb) > sizeof(path)) {
 		if (loglvl & ERRORS) {
 			logentry(clienth, clientp, recvc,
 				"path truncation occurred");
@@ -267,30 +285,41 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 	 *	$args = $rest_of_path + "?" + $args
 	 */
 	if (stat(path, &dir) == -1) {
-		memmove(argsc, args, strlen(args));
+		memmove(traversec, traverse, strlen(traverse));
 		snprintf(path, sizeof(path), "%s", base);
-		recvbp = recvb + 1;
+		recvbp = recvb;
+
+		/*
+		 * Walk into the selector until some directory or file
+		 * does not exist. Then reconstruct the args, selector
+		 * etc.
+		 */
 		while (recvbp != NULL) {
-			sep = strsep(&recvbp, "/");
+			/* Traverse multiple empty / in selector. */
+			while(recvbp[0] == '/')
+				recvbp++;
+			sep = strchr(recvbp, '/');
+			if (sep != NULL)
+				*sep++ = '\0';
+
 			snprintf(path+strlen(path), sizeof(path)-strlen(path),
-				"/%s", sep);
+				"/%s", recvbp);
+			/* path is now always at least '/' */
 			if (stat(path, &dir) == -1) {
-				c = strrchr(path, '/');
-				if (c != NULL) {
-					*c++ = '\0';
-					snprintf(args, sizeof(args),
-						"/%s%s%s%s%s",
-						c,
-						(recvbp != NULL)? "/" : "",
-						(recvbp != NULL)? recvbp : "",
-						(argsc[0] != '\0')? "?" : "",
-						(argsc[0] != '\0')? argsc : ""
-					);
-				}
+				path[strlen(path)-strlen(recvbp)-1] = '\0';
+				snprintf(traverse, sizeof(traverse),
+					"/%s%s%s%s",
+					recvbp,
+					(sep != NULL)? "/" : "",
+					(sep != NULL)? sep : "",
+					(traversec[0] != '\0')? traversec : ""
+				);
 				/* path fallthrough */
 				pathfallthrough = 1;
 				break;
 			}
+			/* Append found directory to path. */
+			recvbp = sep;
 		}
 	}
 
@@ -310,7 +339,8 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		if (S_ISDIR(dir.st_mode)) {
 			for (i = 0; i < sizeof(indexf)/sizeof(indexf[0]);
 					i++) {
-				if (strlen(path) + strlen(indexf[i])
+				len = strlen(path);
+				if (len + strlen(indexf[i]) + ((path[len-1] == '/')? 0 : 1)
 						>= sizeof(path)) {
 					if (loglvl & ERRORS) {
 						logentry(clienth, clientp,
@@ -319,12 +349,26 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 					}
 					return;
 				}
-				strncat(path, indexf[i],
-						sizeof(path)-strlen(path)-1);
+				/*
+				 * The size check for strcat to work is
+				 * calculated above this comment.
+				 *
+				 * Until strlcat isn't properly in all
+				 * linux libcs, we keep to this. OpenBSD
+				 * will complain about strcat and
+				 * smart-ass gcc will cmplain about
+				 * strncat of one char static char array
+				 * is an overflow.
+				 */
+				if (path[len-1] != '/')
+					strcat(path, "/");
+				strcat(path, indexf[i]);
 				fd = open(path, O_RDONLY);
 				if (fd >= 0)
 					break;
-				path[strlen(path)-strlen(indexf[i])] = '\0';
+
+				/* Not found. Clear path from indexf. */
+				path[len] = '\0';
 			}
 		} else {
 			fd = open(path, O_RDONLY);
@@ -339,10 +383,9 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		}
 	}
 
+	/* Some file was opened. Serve it. */
 	if (fd >= 0) {
 		close(fd);
-		if (loglvl & FILES)
-			logentry(clienth, clientp, recvc, "serving");
 
 		c = strrchr(path, '/');
 		if (c == NULL)
@@ -356,8 +399,10 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 		if (pathfallthrough &&
 				!(type->f == handledcgi || type->f == handlecgi)) {
 			dprintf(sock, notfounderr, recvc);
-			if (loglvl & ERRORS)
-				logentry(clienth, clientp, recvc, "not found");
+			if (loglvl & ERRORS) {
+				logentry(clienth, clientp, recvc,
+					"handler in path fallthrough not allowed");
+			}
 			return;
 		}
 
@@ -366,17 +411,25 @@ handlerequest(int sock, char *req, int rlen, char *base, char *ohost,
 			if (loglvl & ERRORS)
 				logentry(clienth, clientp, recvc, "nocgi error");
 		} else {
+			if (loglvl & FILES)
+				logentry(clienth, clientp, recvc, "serving");
+
 			type->f(sock, path, port, base, args, sear, ohost,
-				clienth, serverh, istls);
+				clienth, serverh, istls, recvc, traverse);
 		}
 	} else {
-		/*
-		 * If we had to traverse the path, do not allow directory
-		 * listings, only dynamic content.
-		 */
+		if (pathfallthrough && S_ISDIR(dir.st_mode)) {
+			dprintf(sock, notfounderr, recvc);
+			if (loglvl & ERRORS) {
+				logentry(clienth, clientp, recvc,
+					"directory listing in traversal not allowed");
+			}
+			return;
+		}
+
 		if (!pathfallthrough && S_ISDIR(dir.st_mode)) {
 			handledir(sock, path, port, base, args, sear, ohost,
-				clienth, serverh, istls);
+				clienth, serverh, istls, recvc, traverse);
 			if (loglvl & DIRS) {
 				logentry(clienth, clientp, recvc,
 							"dir listing");
@@ -405,7 +458,6 @@ sighandler(int sig)
 	case SIGQUIT:
 	case SIGABRT:
 	case SIGTERM:
-	case SIGKILL:
 		if (dosyslog) {
 			closelog();
 		} else if (logfile != NULL && glfd != -1) {
@@ -434,7 +486,6 @@ initsignals(void)
 	signal(SIGQUIT, sighandler);
 	signal(SIGABRT, sighandler);
 	signal(SIGTERM, sighandler);
-	signal(SIGKILL, sighandler);
 
 	signal(SIGPIPE, SIG_IGN);
 }
@@ -541,15 +592,14 @@ main(int argc, char *argv[])
 {
 	struct addrinfo hints;
 	struct sockaddr_storage clt, slt;
-	struct linger lingerie;
 	socklen_t cltlen, sltlen;
 	int sock, dofork = 1, inetf = AF_UNSPEC, usechroot = 0,
 	    nocgi = 0, errno_save, nbindips = 0, i, j,
 	    nlfdret, *lfdret, listfd, maxlfd, istls = 0,
 	    dotls = 0, dohaproxy = 0, tcpver = -1, haret = 0,
 #ifdef ENABLE_TLS
-	    tlspipe[2], shufbuf[1025],
-	    shuflen, wlen, shufpos,
+	    tlssocks[2], shufbuf[1025],
+	    shuflen, wlen, shufpos, tlsclientreader,
 #endif /* ENABLE_TLS */
 	    maxrecv, retl,
 	    rlen = 0;
@@ -608,7 +658,7 @@ main(int argc, char *argv[])
 		logfile = EARGF(usage());
 		break;
 	case 'n':
-		revlookup = 0;
+		revlookup = 1;
 		break;
 	case 'o':
 		sport = EARGF(usage());
@@ -788,7 +838,7 @@ main(int argc, char *argv[])
 			perror("chdir");
 			return 1;
 		}
-		base = "/";
+		base = "";
 		if (chroot(".") < 0) {
 			perror("chroot");
 			return 1;
@@ -798,9 +848,9 @@ main(int argc, char *argv[])
 		return 1;
 	}
 
-	/* strip / at the end, except if it is "/" */
-	for (p = base + strlen(base); p > base + 1 && p[-1] == '/'; --p)
-		p[-1] = '\0';
+	/* strip / at the end of base */
+	for (p = base + strlen(base) - 1; p >= base && p[0] == '/'; --p)
+		p[0] = '\0';
 
 	if (dropprivileges(gr, us) < 0) {
 		perror("dropprivileges");
@@ -907,9 +957,10 @@ main(int argc, char *argv[])
 
 #ifdef __OpenBSD__
 			snprintf(promises, sizeof(promises),
-			         "rpath inet stdio %s %s",
-			         nocgi     ? ""    : "proc exec",
-			         revlookup ? "dns" : "");
+			         "rpath inet stdio %s %s %s",
+			         !nocgi || dotls ? "proc" : "",
+			         nocgi           ? ""     : "exec",
+			         revlookup       ? "dns"  : "");
 			if (pledge(promises, NULL) == -1) {
 				perror("pledge");
 				exit(1);
@@ -933,7 +984,12 @@ read_selector_again:
 				istls = 1;
 				if (tls_accept_socket(tlsctx, &tlsclientctx, sock) < 0)
 					return 1;
-				if (tls_handshake(tlsclientctx) < 0)
+				wlen = TLS_WANT_POLLIN;
+				while (wlen == TLS_WANT_POLLIN \
+						|| wlen == TLS_WANT_POLLOUT) {
+					wlen = tls_handshake(tlsclientctx);
+				}
+				if (wlen == -1)
 					return 1;
 			}
 #endif /* ENABLE_TLS */
@@ -1032,41 +1088,95 @@ read_selector_again:
 
 #ifdef ENABLE_TLS
 			if (istls) {
-				if (pipe(tlspipe) < 0) {
-					perror("tls_pipe");
+				if (socketpair(AF_LOCAL, SOCK_STREAM, 0, tlssocks) < 0) {
+					perror("tls_socketpair");
 					return 1;
 				}
 
 				switch(fork()) {
 				case 0:
-					sock = tlspipe[1];
-					close(tlspipe[0]);
+					sock = tlssocks[1];
+					close(tlssocks[0]);
 					break;
 				case -1:
 					perror("fork");
 					return 1;
 				default:
-					close(tlspipe[1]);
+					tlsclientreader = 1;
+					switch(fork()) {
+					case 0:
+						break;
+					case -1:
+						perror("fork");
+						return 1;
+					default:
+						tlsclientreader = 0;
+					}
+
+					close(tlssocks[tlsclientreader? 1 : 0]);
 					do {
-						shuflen = read(tlspipe[0], shufbuf, sizeof(shufbuf)-1);
-						if (shuflen == -1 && errno == EINTR)
-							continue;
-						for (shufpos = 0; shufpos < shuflen; shufpos += wlen) {
-							wlen = tls_write(tlsclientctx, shufbuf+shufpos, shuflen-shufpos);
-							if (wlen < 0) {
-								fprintf(stderr, "tls_write failed: %s\n", tls_error(tlsclientctx));
-								return 1;
+						if (tlsclientreader) {
+							shuflen = read(tlssocks[0],
+								shufbuf,
+								sizeof(shufbuf)-1);
+						} else {
+							shuflen = tls_read(tlsclientctx,
+								shufbuf,
+								sizeof(shufbuf)-1);
+							if (shuflen == TLS_WANT_POLLIN \
+									|| shuflen == TLS_WANT_POLLOUT) {
+								continue;
 							}
 						}
-					} while(shuflen > 0);
+						if (shuflen == -1 && errno == EINTR)
+							continue;
+						for (shufpos = 0; shufpos < shuflen;
+								shufpos += wlen) {
+							if (tlsclientreader) {
+								wlen = tls_write(tlsclientctx,
+									shufbuf+shufpos,
+									shuflen-shufpos);
+								if (wlen == TLS_WANT_POLLIN
+									|| wlen == TLS_WANT_POLLOUT) {
+									wlen = 0;
+									continue;
+								}
+								if (wlen < 0) {
+									fprintf(stderr,
+										"tls_write failed: %s\n",
+										tls_error(tlsclientctx));
+									return 1;
+								}
+							} else {
+								wlen = write(tlssocks[1],
+									shufbuf+shufpos,
+									shuflen-shufpos);
+								if (wlen < 0) {
+									perror("write");
+									return 1;
+								}
+							}
+						}
+					} while (shuflen > 0);
 
-					tls_close(tlsclientctx);
-					tls_free(tlsclientctx);
-					close(tlspipe[0]);
+					if (tlsclientreader) {
+						wlen = TLS_WANT_POLLIN;
+						while (wlen == TLS_WANT_POLLIN \
+								|| wlen == TLS_WANT_POLLOUT) {
+							wlen = tls_close(tlsclientctx);
+						}
+						tls_free(tlsclientctx);
+					}
 
-					waitforpendingbytes(sock);
-					shutdown(sock, SHUT_RDWR);
-					close(sock);
+					lingersock(tlssocks[tlsclientreader? 0 : 1]);
+					shutdown(tlssocks[tlsclientreader? 0 : 1],
+							tlsclientreader? SHUT_WR : SHUT_RD);
+					close(tlssocks[tlsclientreader? 0 : 1]);
+
+					if (tlsclientreader) {
+						lingersock(sock);
+						close(sock);
+					}
 					return 0;
 				}
 			}
@@ -1078,30 +1188,8 @@ read_selector_again:
 					clienth, clientp, serverh, serverp,
 					nocgi, istls);
 
-			if (!istls) {
-				/*
-				 * On close only wait for at maximum 60
-				 * seconds for all data to be transmitted
-				 * before forcefully closing the
-				 * connection.
-				 */
-				lingerie.l_onoff = 1;
-				lingerie.l_linger = 60;
-				setsockopt(sock, SOL_SOCKET, SO_LINGER,
-						&lingerie, sizeof(lingerie));
-				/*
-				 * Force explict flush of buffers using
-				 * TCP_NODELAY.
-				 */
-				j = 1;
-				setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
-						&j, sizeof(int));
-				waitforpendingbytes(sock);
-				j = 0;
-				setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
-						&j, sizeof(int));
-				shutdown(sock, SHUT_RDWR);
-			}
+			lingersock(sock);
+			shutdown(sock, SHUT_RDWR);
 			close(sock);
 
 			if (loglvl & CONN) {
